open Ast


let rec linearize fg expr context =
  let (expr_desc, expr_id) = expr in
    match expr_desc with
      | Eident x -> 
	  context fg (mkident x)
      | Econstant const ->
	  let (fg, fresh) = Fresh.nextvar fg in
	    mklet fresh expr (context fg (mkident fresh))
      | Elet(var, bound, body) ->
	  linearize fg bound (fun fg boundv ->
	    mklet var boundv (linearize fg body context)
	  )
      | Efunction(var, body) ->
	  let body        = linearize fg body (fun _ x -> x) in
	  let (fg, fresh) = Fresh.nextvar fg in
	    mklet fresh (mkfun var body) (context fg (mkident fresh))
      | Efixpoint(vfix, var, body) ->
	  let body        = linearize fg body (fun _ x -> x) in
	  let (fg, fresh) = Fresh.nextvar fg in
	    mklet fresh (mkfix vfix var body) (context fg (mkident fresh))
      | Eapply(func, args) ->
	  let (fg, fresh) = Fresh.nextvar fg in
	  linearize fg func (fun fg func_ident ->
	    linearize_apply fg args [] (fun fg args ->
	      let let_bound = mkexpr (Eapply(func_ident, args)) in
		mklet fresh let_bound (context fg (mkident fresh))))
      | Ematch(matched_expr, matching) ->
	  let (fg, fresh) = Fresh.nextvar fg in
	  let matching = List.map (fun (patt, action) ->
	    (patt, linearize fg action (fun _ x -> x))
	  ) matching in
	    linearize fg matched_expr (fun fg x ->
	      mklet fresh (mkmatch x matching) (context fg (mkident fresh))
	    )
      | Etuple fields ->
	  let (fg, fresh) = Fresh.nextvar fg in
	    linearize_tuple fg fields [] (fun fg fields ->
	      mklet fresh (mktuple fields) (context fg (mkident fresh))
	    )
      | Econstruct(cons, None) ->
	  let (fg, fresh) = Fresh.nextvar fg in
	    mklet fresh expr (context fg (mkident fresh))	  
      | Econstruct(cons, Some data) ->
	  let (fg, fresh) = Fresh.nextvar fg in
	    linearize fg data (fun fg x ->
	      mklet fresh (mkconstruct cons (Some x))
		(context fg (mkident fresh)))

	    
and linearize_apply fg args acc context =
  match args with
    | [] ->
	context fg (List.rev acc)
    | a :: tl ->
	linearize fg a (fun fg a_ident ->
	  linearize_apply fg tl (a_ident :: acc) context
	)

and linearize_tuple fg fields acc context =	  
  match fields with
    | [] ->
	context fg (List.rev acc)
    | a :: tl ->
	linearize fg a (fun fg a_ident ->
	  linearize_tuple fg tl (a_ident :: acc) context
	)
		    
let perform_linearization (decls, main) fg =
  (decls, linearize fg main (fun _ x -> x))
